import os
from arjuna.interact.gui.gom.impl.base import BaseGui
from arjuna.tpi.enums import ArjunaOption

class DefaultGui(BaseGui):
    '''
        This class provides a means for direct creation of Gui objects without inheritance.
    '''

    def __init__(self, automator, ns_dir, def_file_name, parent=None):
        super().__init__(automator, ns_dir=ns_dir, def_file_name=def_file_name, parent=parent)


class SimpleBaseGui(BaseGui):
    '''
        This class is meant to be inherited from, so that you can implement a more involed GOM model.
    '''

    def __init__(self, automator, ns_dir):
        super().__init__(automator, ns_dir=ns_dir)

class SimpleBaseWidget(BaseGui):

    def __init__(self, automator, parent, ns_dir):
        super().__init__(automator, parent=parent, ns_dir=ns_dir)
        label = self.__class__.__name__


from enum import Enum, auto

class PartialActionType(Enum):
    CLICK = auto()
    CLICK_AND_HOLD = auto()
    MOUSE_UP = auto()
    RIGHT_CLICK = auto()
    DOUBLE_CLICK = auto()
    DRAG_AND_DROP = auto()
    SEND_KEYS = auto()
    MOVE_TO = auto()
    PAUSE = auto()

class DefaultCompositeAction:

    def __init__(self, automator, partial_actions):
        self.__partial_actions = partial_actions
        self.__automator =  automator

    def perform(self):
        self.__automator.perform_composite_action(self)

    @property
    def partial_actions(self):
        return self.__actions


class PartialAction:

    def __init__(self, action_type):
        self.__type = action_type
        self.__args = dict()

    def add_target_gui_element(self, element):
        args["target_element"] = element

    def add_source_element(self, element):
        args["source_element"] = element

    def add_target_point(self, point):
        args["target_point"] = point

    def add_source_point(self, point):
        args["source_point"] = point

    def add_offset(self, offset):
        args["offset"] = offset

    def add_key_chord(self, chord):
        args["chord"] = chord

    def add_pause_time(self, t):
        args["pause"] = t

    @property
    def action_type(self):
        return self.__type

    @property
    def args(self):
        return self.__args


class SingleActionChain:

    def __init__(self, automator):
        self.__automator = automator
        self.__actions = list()

    def __default_action_chain(partial_action_type):
        action = PartialAction(partial_action_type)
        self.__actions.append(action)

    def __default_point_action(partial_action_type, point):
        action = PartialAction(partial_action_type)
        action.add_target_point(point)
        self.__actions.append(action)
        return self

    def __default_element_action(partial_action_type, element):
        action = PartialAction(partial_action_type)
        action.add_target_guielement(element)
        self.__actions.append(action)
        return self

    def __click_action(self, *, atype, point, element):
        if point is None:
            if element is None:
                return self.__default_action_chain(PartialActionType.DOUBLE_CLICK)
            else:
                self.__default_element_action(PartialActionType.DOUBLE_CLICK, element)
        else:
            if element is None:
                return self.__default_point_action(PartialActionType.DOUBLE_CLICK, point)
            else:
                raise Exception("You can not provide both point and element to {} action.".format(atype))

    def release(self):
        return self.__default_action_chain(PartialActionType.MOUSE_UP)

    def click(self, *, point=None, element=None):
        return self.__click_action(PartialActionType.CLICK, point, element)

    def press(self, *, point=None, element=None):
        return self.__click_action(PartialActionType.CLICK_AND_HOLD, point, element)

    def right_click(self, *, point=None, element=None):
        return self.__click_action(PartialActionType.RIGHT_CLICK, point, element)

    def double_click(self, *, point=None, element=None):
        return self.__click_action(PartialActionType.DOUBLE_CLICK, point, element)

    def drag_and_drop(self, *, source_element=None, source_point=None, dest_element=None, dest_point=None, offset=None):
        action = PartialAction(PartialActionType.DRAG_AND_DROP)
        if source_element is not None: action.add_source_element(source_element)
        if source_point is not None: action.add_source_point(source_point)
        if dest_element is not None: action.add_target_point(dest)
        if offset is not None: action.add_offset(xy)
        if dest_point is not None: action.add_target_point(dest_point)
        self.__actions.append(action)
        return self

    def send_text(self, text, *modifier_keys):
        return self

    def send_keys(self, key_chord=None):
        return self

    def hover(self, *, dest_element=None, dest_point=None, offset=None):
        action = PartialAction(PartialActionType.MOVE_TO)
        if dest is not None: action.add_target_gui_element(dest_element)
        if offset is not None: action.add_offset(offset)
        if dest_point is not None: action.add_target_point(dest_point)
        self.__actions.append(action)

    def pause(self, duration):
        action = PartialAction(PartialActionType.PAUSE)
        action.add_pause_time(duration)
        self.__actions.append(action)
        return self

    def build(self):
        return DefaultCompositeAction(self.__automator, self.__actions)

    def perform(self):
        action = this.build()
        action.perform()


class SingleActionChainBuilder:

    def chain(self):
        pass

    def build(self):
        pass

    def perform(self):
        pass


from arjuna.tpi.guiauto.helpers import _WithType
from .guidef import GuiDef

class Gui:

    def __init__(self, gui_mgr, automator, gui_def, parent=None):
        self.__guimgr = gui_mgr
        self.__guidef = gui_def
        self.__automator = automator
        self.__parent = parent

    @property
    def gui_def(self):
        return self.__guidef

    @property
    def automator(self):
        return self.__automator

    def define_gui(self, automator, label=None, name=None, qual_name=None, def_file_path=None):
        gui_def = GuiDef(self.__guimgr.name_store, self.__guimgr.namespace_dir, automator, label, def_file_path)
        gui = Gui(self, automator, gui_def, parent=self)
        return gui

    def __lmd(self, *locators):
        out = []
        for locator in locators:
            if locator.wtype == _WithType.GNS_NAME:
                out.extend(self.gui_def.convert_to_with(locator))
            else:
                out.append(locator)
        return out

    def define_element(self, *with_locators):
        return self.automator.element(*self.__lmd(*with_locators))

    def define_multielement(self, *with_locators):
        return self.automator.multielement(*self.__lmd(*with_locators))

    def define_dropdown(self, *with_locators):
        return self.automator.dropdown(*self.__lmd(*with_locators))

    def define_radiogroup(self, *with_locators):
        return self.automator.radiogroup(*self.__lmd(*with_locators))

    def define_frame(self, *with_locators):
        return self.automator.frame(*self.__lmd(*with_locators))

    @property
    def alert(self):
        return self.automator.alert

    @property
    def main_window(self):
        return self.automator.main_window

    @property
    def browser(self):
        return self.automator.browser

    def set_slomo(self, on, interval=None):
        self.automator.set_slomo(on, interval)



class SourceContent:

    def __init__(self, src):
        self.__src = src

    @property
    def src(self):
        return self.__src

    @property
    def root(self):
        return self.src.get_root_content()

    @property
    def all(self):
        return self.src.get_full_content()

    @property
    def inner(self):
        return self.src.get_inner_content()

    @property
    def text(self):
        return self.src.get_text_content()

class DefaultGuiSource:

    def __init__(self, automator, src):
        self.__automator = automator
        self.__content =  SourceContent(src)

    # @staticmethod
    # def define_source(self, automator, *args):
    #     return DefaultGuiSource(automator, args)

    @property
    def content(self):
        return self.__content





from enum import Enum, auto
import random

from arjuna.interact.gui.auto.invoker.source import DefaultGuiSource

class GuiComponentType(Enum):
    ELEMENT = auto()
    MULTI_ELEMENT = auto()
    DROPDOWN = auto()
    RADIOGROUP = auto()
    BROWSER = auto()
    DOMROOT = auto()
    FRAME = auto()
    WINDOW = auto()
    MAIN_WINDOW = auto()
    CHILD_WINDOW = auto()
    ALERT = auto()

class GuiAutoComponentFactory:

    @staticmethod
    def Element(automator, impl_element):
        return DefaultGuiElement(automator, impl_element)

    @staticmethod
    def MultiElement(automator, impl_melement):
        return DefaultGuiMultiElement(automator, impl_melement)

    @staticmethod
    def DropDown(automator, impl_dropdown):
        return DefaultDropDown(automator, impl_dropdown)

    @staticmethod
    def RadioGroup(automator, impl_radiogroup):
        return DefaultRadioGroup(automator, impl_radiogroup)

    @staticmethod
    def Alert(automator, impl_alert):
        return DefaultAlert(automator, impl_alert)

    @staticmethod
    def MainWindow(automator, impl_main_window):
        return DefaultMainWindow(automator, impl_main_window)

    @staticmethod
    def DomRoot(automator, impl_dom_root):
        return DefaultDomRoot(automator, impl_dom_root)

    @staticmethod
    def Frame(automator, impl_frame):
        return DefaultFrame(automator, impl_frame)

    @staticmethod
    def Browser(automator, impl_browser):
        return DefaultBrowser(automator, impl_browser)


class BaseComponent:

    def __init__(self, automator, component_type, impl_component):
        self.__automator = automator
        self.__comp_type = component_type
        self.__impl_component = impl_component

    @property
    def _automator(self):
        return self.__automator

    @property
    def _test_session(self):
        return self.automator.test_session

    @property
    def component_type(self):
        return self.__comp_type

    @property
    def impl(self):
        return self.__impl_component

    @property
    def source(self):
        return DefaultGuiSource(self._automator, self.impl.get_source())

    def _emd(self, *locators):
        return self._automator._create_lmd(*locators)

    def configure(self, config):
        self.impl.configure(config.settings)
        return self

class DefaultGuiElement(BaseComponent):

    def __init__(self, automator, impl, index=None):
        super().__init__(automator, GuiComponentType.ELEMENT, impl)
        self.__index = index

    def enter_text(self, text):
        self.impl.enter_text(text)

    def set_text(self, text):
        self.impl.set_text(text)

    def click(self):
        self.impl.click()

    def wait_until_present(self):
        self.impl.wait_until_present()

    def wait_until_visible(self):
        self.impl.wait_until_visible()

    def wait_until_clickable(self):
        self.impl.wait_until_clickable()

    def check(self):
        self.impl.check()

    def uncheck(self):
        self.impl.uncheck()

    def identify(self):
        self.impl.identify()

class DefaultGuiMultiElement(BaseComponent):

    def __init__(self, automator, impl):
        super().__init__(automator, GuiComponentType.ELEMENT, impl)

    def at_index(self, index):
        return DefaultGuiElement(self._automator, self.impl.get_instance_at_index(index), index)

    @property
    def first(self):
        return self.at_index(0)

    @property
    def last(self):
        return self.at_index(self.length -1)

    @property
    def random(self):
        return self.at_index(random.randint(0, self.length-1))

    @property
    def length(self):
        return self.impl.get_instance_count()

class MultiElementSelectable(BaseComponent):

    def __init__(self, automator, comp_type, impl):
        super().__init__(automator, comp_type, impl)

    def has_value_selected(self, value):
        return self.impl.has_value_selected(value)

    def has_index_selected(self, index):
        return self.impl.has_index_selected(index)

    def select_value(self, value):
        self.impl.select_by_value(value)

    def select_index(self, index):
        self.impl.select_by_index(index)

    @property
    def first_selected_option_value(self):
        return self.impl.get_first_selected_option_value()

class DefaultDropDown(MultiElementSelectable):

    def __init__(self, automator, impl):
        super().__init__(automator, GuiComponentType.DROPDOWN, impl)

    @property
    def _option_locators(self):
        pass

    @_option_locators.setter
    def option_locators(self, *with_locators):
        self.impl.set_option_locators(self.__emd(*with_locators))

    @property
    def _option_container(self):
        pass

    @_option_container.setter
    def option_container(self, *with_locators):
        self.impl.set_option_container(self.__emd(*with_locators))

    @property
    def first_selected_option_text(self):
        return self.impl.get_first_selected_option_text()

    def has_visible_text_selected(self, text):
        return self.impl.has_visible_text_selected(text)

    def select_visible_text(self, text):
        self.impl.select_by_visible_text(text)

    def send_option_text(self, text):
        self.impl.send_option_text(text)

class DefaultRadioGroup(MultiElementSelectable):

    def __init__(self, automator, impl):
        super().__init__(automator, GuiComponentType.RADIOGROUP, impl)

    def configure(self, config):
        pass

class DefaultAlert(BaseComponent):

    def __init__(self, automator, impl):
        super().__init__(automator, GuiComponentType.ALERT, impl)

    def confirm(self):
        self.impl.confirm()

    def dismiss(self):
        self.impl.dismiss()

    @property
    def text(self):
        return self.impl.get_text()

    def send_text(self, text):
        self.impl.send_text(text)


class DefaultBrowser(BaseComponent):

    def __init__(self, automator, impl):
        super().__init__(automator, GuiComponentType.BROWSER, impl)

    def go_to_url(self, url):
        self.impl.go_to_url(url)

    def go_back(self):
        self.impl.go_back()

    def go_forward(self):
        self.impl.go_forward()
        self._send_request(ArjunaComponent.GUI_AUTOMATOR, GuiAutoActionType.BROWSER_GO_FORWARD)

    def refresh(self):
        self.impl.refresh()

    @property
    def dom_root(self):
        return DefaultDomRoot(self._automator, self.impl.dom_root)

class BaseFrame(BaseComponent):

    def __init__(self, automator, comp_type, impl):
        super().__init__(automator, comp_type, impl)    
        
    def focus(self):
        self.impl.focus()

    def frame(self, *with_locators):
        return GuiAutoComponentFactory.Frame(self._automator, self.impl.define_frame(self._emd(*with_locators)))

    @property
    def parent(self):
        return GuiAutoComponentFactory.Frame(self._automator, self.impl.parent)

    def enumerate_frames(self):
        return self.impl.enumerate_frames()

class DefaultFrame(BaseFrame):

    def __init__(self, automator, impl):
        super().__init__(automator, GuiComponentType.FRAME, impl)

class DefaultDomRoot(BaseFrame):

    def __init__(self, automator, impl):
        super().__init__(automator, GuiComponentType.DOMROOT, impl)
    
    @property
    def parent(self):
        raise Exception("DOM root does not have a parent frame.")


class AbstractBasicWindow(BaseComponent):

    def __init__(self, automator, comp_type, impl):
        super().__init__(automator, comp_type, impl)

    @property
    def title(self):
        return self.impl.title

    def focus(self):
        self.impl.focus()


class DefaultChildWindow(AbstractBasicWindow):

    def __init__(self, automator, impl):
        super().__init__(automator, GuiComponentType.CHILD_WINDOW, impl)

    def close(self):
        self.impl.close()

    @property
    def main_window(self):
        self._automator.main_window


class DefaultMainWindow(AbstractBasicWindow):

    def __init__(self, automator, impl):
        super().__init__(automator, GuiComponentType.MAIN_WINDOW, impl)

    def maximize(self):
        self.impl.maximize()

    def _take_element_finding_action(self, setu_action_type, *setu_args):
        response = self._send_request(ArjunaComponent.GUI_AUTOMATOR, setu_action_type, *setu_args)
        return response.get_value_for_gui_component_setu_id()

    def child_window(self, *with_locators):
        win =  self.impl.define_child_window(self._emd(*with_locators))
        return DefaultChildWindow(self._automator, win)

    @property
    def latest_child_window(self):
        return DefaultChildWindow(self._automator, self.impl.get_latest_child_window())
 
    def close_all_child_windows(self):
        self.impl.close_all_child_windows()



from .base_element import BaseElement, ElementConfig
from arjuna.interact.gui.auto.impl.source.parser import ElementXMLSourceParser

class GuiElement(BaseElement, ElementConfig):
    
    def __init__(self, automator, emd, parent=None):
        BaseElement.__init__(self, automator, emd, parent, obj_name="GuiElement")
        ElementConfig.__init__(self, automator)
        from .element_conditions import GuiElementConditions
        self.__conditions_handler = GuiElementConditions(self)
        self.__source_parser = None
        self.__automator = automator

    def __get_attr_value_from_remote(self, attr, optional=False):
        return self.__return_attr_value(self.dispatcher.get_attr_value(attr, optional))

    def get_source_from_remote(self):
        return self.__get_attr_value_from_remote("outerHTML")

    def load_source_parser(self):
        raw_source = self.get_source_from_remote()
        if self.__source_parser is None:
            self.__source_parser = ElementXMLSourceParser(self)
        self.__source_parser.load()

    def find(self):
        self.parent_container.find_element(self)

    def find_without_wait(self):
        self.parent_container.find_element_without_wait(self)

    identify = find

    #Override
    def find_if_not_found(self):
        if not self.is_found():
            self.find()

    def _is_partial_element(self):
        return False

    def _get_instance_number(self):
        raise Exception("Instance number is applicable only for partial gui elements.")

    def __append_instance_number(self, d):
        if self._is_partial_element():
            d["isInstanceAction"] = True
            d["instanceIndex"] = self._get_instance_number()
        return d

    def _kwargs(self, **kwargs):
        return self.__append_instance_number(kwargs)

    def _noargs(self):
        return self.__append_instance_number({})

    def _only_send_text(self, text):
        self.dispatcher.send_text(text)

    def _only_click(self):
        self.dispatcher.click()

    def __return_attr_value(self, result):
        return result and result or None

    def click(self):
        self.find_if_not_found()
        self.__wait_until_clickable_if_configured()
        self._only_click()

    def __conditional_selected_state_click(self, condition_state):
        self.find_if_not_found()
        selected = self.is_selected()
        if selected == condition_state:
            self.__wait_until_clickable_if_configured()
            self._only_click()

    def select(self):
        self.__conditional_selected_state_click(False)

    def deselect(self):
        self.__conditional_selected_state_click(True)

    wait_until_present = find

    def wait_until_visible(self):
        self.find_if_not_found()
        self.conditions.IsVisible().wait()

    def wait_until_clickable(self):
        self.find_if_not_found()
        self.conditions.IsClickable().wait()

    def wait_until_selected(self):
        self.find_if_not_found()
        self.conditions.IsVisible().wait()

    #################################
    ### State Checking
    #################################
    def is_selected(self):
        self.find_if_not_found()
        return self.dispatcher.is_selected()

    def is_visible(self):
        self.find_if_not_found()
        return self.dispatcher.is_visible()

    def is_clickable(self):
        self.find_if_not_found()
        return self.dispatcher.is_clickable()

    @property
    def conditions(self):
        return self.__conditions_handler

    def _only_clear_text(self):
        self.dispatcher.clear_text()

    def _only_enter_text(self, text):
        self._only_send_text(text)

    #################################
    ### Textbox abstraction
    #################################

    def clear_text(self):
        self.find_if_not_found()
        self.__wait_until_clickable_if_configured()
        self._only_click()
        self._only_clear_text()

    def send_text(self, text):
        self.find_if_not_found()
        self.__wait_until_clickable_if_configured()
        self._only_enter_text(text)

    def enter_text(self, text):
        self.find_if_not_found()
        self.__wait_until_clickable_if_configured()
        self._only_click()
        self._only_enter_text(text)

    def set_text(self, text):
        self.find_if_not_found()
        self.__wait_until_clickable_if_configured()
        self._only_click()
        self._only_clear_text()
        self._only_enter_text(text)

    def has_entered_text(self, text):
        pass

    def has_entered_partial_text(self, text):
        pass

    #################################
    ### Checkbox abstraction
    #################################
    def check(self):
        self.select()

    def uncheck(self):
        self.deselect()

    def toggle_checkbox(self):
        self.click()

    def is_checked(self):
        return self.is_selected()

    def __wait_until_clickable_if_configured(self):
        if self._should_check_pre_state(): self.wait_until_clickable()

    def get_source(self, refind=True, reload=True):
        if refind:
            self.find_if_not_found()
        if reload:
            self.load_source_parser()
        return self.__source_parser

    def get_property_value(self, attr):
        self.find_if_not_found()
        return self.__get_attr_value_from_remote(attr)

    def perform_action_chain(self, single_action_chain):
        from arjuna.interact.gui.auto.automator.actions import SingleActionChain
        action_chain = SingleActionChain(self.__automator, element=self)
        action_chain.perform(single_action_chain)

    def find_element_with_js(self, js):
        raise Exception("With.JS is currently not supported for nested element finding.")

    def find_multielement_with_js(self, js):
        raise Exception("With.JS is currently not supported for nested element finding.")




import re
import os

from enum import Enum, auto
from abc import abstractmethod

from arjuna.core.enums import GuiAutomationContext
from arjuna.interact.gui.auto.impl.locator.emd import GuiElementMetaData, Locator, ImplWith
from arjuna.tpi.guiauto.helpers import With
from arjuna.tpi.enums import ArjunaOption

class FileFormat(Enum):
    GNS = auto()
    MGNS = auto()
    XLS = auto()
    XLSX = auto()

class GuiNamespaceLoaderFactory:

    # Returns GuiNamespaceLoader
    @classmethod
    def create_namespace_loader(cls, config, ns_file_path):
        multi_context_enabled = config.get_arjuna_option_value(ArjunaOption.GUIAUTO_DEF_MULTICONTEXT).as_bool()
        context = multi_context_enabled and None or config.get_guiauto_context()
        _, file_extension = os.path.splitext(ns_file_path)
        ext = file_extension.upper()[1:]
        considered_path = ns_file_path
        try:
            file_format = FileFormat[ext]
        except:
            raise Exception("Unsupported format for namespace: {}".format(file_extension))
        else:
            full_file_path = ns_file_path
            if os.path.isdir(full_file_path):
                raise Exception("Namespace file path is a directory and not a file: {}".format(considered_path))
            elif not os.path.isfile(full_file_path):
                raise Exception("Namespace file path is invalid: {}".format(considered_path))

            if file_format == FileFormat.GNS:
                if multi_context_enabled:
                    return MGNSFileLoader(full_file_path)
                else:
                    return GNSFileLoader(full_file_path, context)
            else:
                raise Exception("Unsupported format for namespace: {}".format(file_extension))


class GuiNamespace:

    def __init__(self, name):
        self.__name = name
        # dict <string, dict<GuiAutomationContext, GuiElementMetaData>>
        self.__ns = {}

    def add_element_meta_data(self, name, context, raw_locators):
        emd = GuiElementMetaData.create_lmd(*raw_locators)
        #emd = GuiElementMetaData(raw_locators, process_args=False)
        name = name.lower()
        if not self.has(name):
            self.__ns[name] = {}
        self.__ns[name][context] = emd

    def has(self, name):
        return name.lower() in self.__ns

    def has_context(self, name, context):
        if self.has(name):
            return context in self.__ns[name.lower()]
        return False

    # Needs to be thread-safe
    # Returns emd for a context for a given gui name
    def get_meta_data(self, name, context):
        if not self.has(name):
            raise Exception("Gui namespace >{}< does not contain element with name: {}".format(self.__name, name))
        elif not self.has_context(name, context):
            raise Exception("Gui namespace >{}< does not contain element with name: {} for context {}".format(self.__name, name, context))
        
        return self.__ns[name.lower()][context]


class BaseGuiNamespaceLoader:

    def __init__(self, name):
        self.__name = name
        self.__namespace = GuiNamespace(name)

    @property
    def name(self):
        return self.__name

    @property
    def namespace(self):
        return self.__namespace

    # Needs to be thread safe
    def add_element_meta_data(self, name, context, locators):
        self.__namespace.add_element_meta_data(name, context, locators)

    def _raise_notafile_exception(self, file_path):
        raise Exception("{} is not a file.".format(file_path))

    def _raise_filenotfound_exception(self, file_path):
        raise Exception("{} is not a valid file path.".format(file_path))

    def _raise_relativepath_exception(self, file_path):
        raise Exception("Gui namespace loader does not accept relative file path. {} is not a full file path.".format(file_path))

class AbstractGNFileLoader(BaseGuiNamespaceLoader):
    NAME_PATTERN = re.compile(r"\[\s*(.*?)\s*\]$")
    LOCATOR_PATTERN = re.compile(r"\s*(.*?)\s*=\s*(.*?)\s*$")

    def __init__(self, ns_file_path):
        super().__init__(os.path.basename(ns_file_path))
        self.__ns_file = None
        self.__ns_path = None
        self.__header_found = False
        self.__last_header = None
        self.__last_auto_contexts = None # list of auto contexts

        #Map<String, Map<GuiAutomationContext,List<Locator>>>
        self.__ns = {}

        if not os.path.isabs(ns_file_path):
            super()._raise_relativepath_exception(ns_file_path)
        elif not os.path.exists(ns_file_path):
            super()._raise_filenotfound_exception(ns_file_path)
        elif not os.path.isfile(ns_file_path):
            super()._raise_notafile_exception(ns_file_path)

        self.__ns_path = ns_file_path
        self.__ns_file = open(self.__ns_path)

    @property
    def header_found(self):
        return self.__header_found

    @header_found.setter
    def header_found(self, flag):
        self.__header_found = flag

    @property
    def last_header(self):
        return self.__last_header

    @last_header.setter
    def last_header(self, name):
        self.__last_header = name

    @property
    def last_auto_contexts(self):
        return self.__last_auto_contexts

    @last_auto_contexts.setter
    def last_auto_contexts(self, contexts):
        self.__last_auto_contexts = contexts

    @property
    def ns_path(self):
        return self.__ns_path

    @property
    def ns_file(self):
        return self.__ns_file

    def load(self):
        for line in self.__ns_file.readlines():
            line = line.strip()
            if not line: 
                continue
            if self._match_header(line):
                self.header_found = True
                self._init_section()
                continue
            else:
                if not self.header_found:
                    raise Exception("Namespace contents must be contained inside a [name] header.")
                else:
                    self._load_section_line(line)
        
        self.__ns_file.close()

        for ename, context_data in self.__ns.items():
            for context, locators in context_data.items():
                self.add_element_meta_data(ename, context, locators)

    def __is_defined(self, name):
        return name.lower() in self.__ns

    def __is_not_first_header(self):
        return self.last_header is not None

    def __validate_duplicate_entry(self, last_name, new_name):
        if (last_name.lower() == new_name.lower()) or self.__is_defined(new_name):
            raise Exception("Found duplicate namespace definition for {} element.".format(new_name))

    def __validate_empty_last_section(self, name):
        if len(self.__ns[name]) == 0:
            raise Exception("Found empty namespace definition for {} element.".format(name))
        else:
            for context, data in self.__ns[name].items():
                if len(data) == 0:
                    raise Exception("Found empty namespace definition for {} context for {} element.".format(context.name, name))


    def _match_header(self, input):
        match = self.NAME_PATTERN.match(input)
        if match:
            current_header = match.group(1)
            if self.__is_not_first_header():
                self.__validate_duplicate_entry(self.last_header, current_header)
                self.__validate_empty_last_section(self.last_header)
            
            # Initialise for new section found
            self.last_header = current_header
            self.last_auto_contexts = None
            self.__ns[self.last_header] = {}
            return True
        else:
            return False

    def _init_contexts_dict(self, name, contexts):
        for context in contexts:
            if context in self.__ns[name]:
                raise Exception("Found duplicate automation context {} in {} namespace definition.".format(context.name, self.last_header))
            else:
                self.__ns[name][context] = []

    @abstractmethod
    def _match_contexts(self, input):
        pass

    @abstractmethod
    def _load_section_line(self, input):
        pass

    @abstractmethod
    def _init_section(self):
        pass

    def _match_locator(self, input):
        match = self.LOCATOR_PATTERN.match(input)
        if match:       
            # locator = Locator(match.group(1), match.group(2), named_args=dict())
            locator = ImplWith(wtype=match.group(1).upper(), wvalue=match.group(2), named_args=dict(), has_content_locator=False)
            # locator = getattr(With, match.group(1).lower())(match.group(2)) # e.g. getattr(With, "ID".lower())("abc")
            if (self.last_auto_contexts is None):
                raise Exception("Locators must be preceded with context information as #context1, context2 construct. Current line: " + input)   

            for context in self.last_auto_contexts:
                self.__ns[self.last_header][context].append(locator)
            return True
        else:
            return False

class GNSFileLoader(AbstractGNFileLoader):

    def __init__(self, ns_file_path, context):
        super().__init__(ns_file_path)
        self.__contexts = [context]

    @property
    def contexts(self):
        return self.__contexts

    def _init_section(self):
        self.last_auto_contexts = self.__contexts
        self._init_contexts_dict(self.last_header, self.contexts)

    def _match_contexts(self, input):
        pass

    def _load_section_line(self, line):
        if self._match_locator(line):
            return
        else:
            raise Exception("Unexpected namespace file entry. Namspace content can either be plaforms or identification definition: " + line)


class MGNSFileLoader(AbstractGNFileLoader):
    PLATFORM_PATTERN = re.compile(r"\s*\#\s*(.*?)\s*$")

    def __init__(self, ns_file_path):
        super().__init__(ns_file_path)

    def _init_section(self):
        pass

    def _match_contexts(self, input):
        match = self.PLATFORM_PATTERN.match(input)
        if match:        
            try:
                raw_contexts = [m.strip().upper() for m in match.group(1).split(",")]
                contexts = [GuiAutomationContext[n] for n in raw_contexts]
            except Exception as e:
                raise Exception("Invalid context name found in header: {}".format(e))
            else:
                self._init_contexts_dict(self.last_header, contexts)
            
            self.last_auto_contexts = contexts
            return True
        else:
            return False

    def _load_section_line(self, line):
        if self._match_contexts(line):
            return
        elif self._match_locator(line):
            return
        else:
            raise Exception("Unexpected namespace file entry. Namspace content can either be plaforms or identification definition: " + line)



from enum import Enum, auto
from arjuna.tpi.guiauto.helpers import With
from arjuna.interact.gui.auto.invoker.mkactions import *
from .component import GuiAutoComponentFactory

from arjuna.interact.gui.auto.impl.automator import GuiAutomator as ImplGuiAutomator
from arjuna.interact.gui.auto.invoker.source import DefaultGuiSource

class GuiAutomatorName(Enum):
    SELENIUM = auto()
    APPIUM = auto()

class GuiAutomator:

    def __init__(self, config=None, extended_config=None):
        from arjuna.tpi import Arjuna
        self.__config = config and config or Arjuna.get_ref_config()
        self.__test_session = self.config.test_session
        self.__dom_root = None
        self.__main_window = None
        self.__browser = None
        self.__auto_context = None
        self._set_impl_automator(ImplGuiAutomator(config, extended_config))
        self._set_main_window(GuiAutoComponentFactory.MainWindow(self, self.impl_automator.main_window))
        self._set_browser(GuiAutoComponentFactory.Browser(self, self.impl_automator.browser))


    def _set_impl_automator(self, automator):
        self.__impl_automator = automator

    @property
    def config(self):
        return self.__config

    @property
    def impl_automator(self):
        return self.__impl_automator

    def _set_automation_context(self, context):
        self.__auto_context = context

    @property
    def context(self):
        return self.__auto_context

    @property
    def test_session(self):
        return self.__test_session

    def is_gui(self):
        return False

    def _create_lmd(self, *locators):
        return self.impl_automator.create_lmd(*locators)

    def element(self, *with_locators):
        impl = self.impl_automator.define_element(self._create_lmd(*with_locators))
        return GuiAutoComponentFactory.Element(self, impl)

    def multi_element(self, *with_locators):
        impl = self.impl_automator.define_multielement(self._create_lmd(*with_locators))
        return GuiAutoComponentFactory.MultiElement(self, impl)

    def dropdown(self, *with_locators, option_container_locators=None, option_locators=None):
        def convert_to_tuple(in_data):
            if isinstance(in_data, With):
                return (in_data, )
            elif type(in_data) in {list, tuple}:
                return in_data
            else:
                raise Exception("Expected a With object or a list/tuple of With objects")
        impl = self.impl_automator.define_dropdown(
            self._create_lmd(*with_locators),
            option_container_lmd= option_container_locators and self._create_lmd(*convert_to_tuple(option_container_locators)) or None,
            option_lmd=option_locators and self._create_lmd(*convert_to_tuple(option_locators)) or None
        )
        return GuiAutoComponentFactory.DropDown(self, impl)

    def radio_group(self, *with_locators):
        impl = self.impl_automator.define_radiogroup(self._create_lmd(*with_locators))
        return GuiAutoComponentFactory.RadioGroup(self, impl)

    def frame(self, *with_locators):
        # Conversion takes place IFrame implementation for locator -> emd
        return self.dom_root.frame(*with_locators)

    def child_window(self, *with_locators):
        return self.main_window.child_window(*with_locators)

    def close_all_child_windows(self):
        self.main_window.close_all_child_windows()

    @property
    def alert(self):
        impl = self.impl_automator.alert_handler.create_alert()
        return GuiAutoComponentFactory.Alert(self, impl)

    @property
    def main_window(self):
        return self.__main_window

    def _set_main_window(self, window):
        self.__main_window = window

    @property
    def latest_child_window(self):
        return self.main_window.latest_child_window

    @property
    def config(self):
        return self.__config

    def _set_config(self, config):
        self.__config = config

    @property
    def dom_root(self):
        return self.browser.dom_root

    def _set_dom_root(self, root):
        self.__dom_root = root

    @property
    def test_session(self):
        return self.__test_session

    @property
    def browser(self):
        return self.__browser

    def _set_browser(self, browser):
        self.__browser = browser

    def enable_slow_motion(self, on, interval=None):
        args = [SetuArg.arg("on", on)]
        if interval:
            args.append(SetuArg.arg("interval", interval))
        self._send_request(GuiAutoActionType.SET_SLOMO, *args)

    def execute_javascript(self, script):
        return self.impl_automator.browser.execute_javascript(script)

    @property
    def source(self):
        return DefaultGuiSource(self, self.impl_automator.get_source())

    def new_action_chain(self):
        return SingleActionChain(self)

    def perform_composite_action(self):
        pass

    def quit(self):
        self.impl_automator.quit()



import abc

from arjuna.interact.gui.auto.impl.base.element_container import ElementContainer
from arjuna.core.enums import GuiActionConfigType

class BaseElement(ElementContainer, metaclass=abc.ABCMeta):
    def __init__(self, automator, emd, parent=None, obj_name=""):
        super().__init__(automator.config, obj_name)
        self.__automator = automator
        self.__parent = parent
        self.__lmd = emd
        self.__found = False
        self.__located_by = None 
        self.__dispatcher_element = None

    @property
    def automator(self):
        return self.__automator

    @property
    def dispatcher(self):
        return self.__dispatcher_element

    @dispatcher.setter
    def dispatcher(self, element):
        self.__dispatcher_element = element

    @property
    def parent_container(self):
        return self.__parent and self.__parent or self.__automator

    def _create_element_flat_or_nested(self, locator_meta_data):
        from arjuna.interact.gui.auto.impl.element.guielement import GuiElement
        return GuiElement(self.__automator, locator_meta_data, parent=self) 

    def _create_multielement_flat_or_nested(self, locator_meta_data):
        from arjuna.interact.gui.auto.impl.element.multielement import GuiMultiElement
        return GuiMultiElement(self.__automator, locator_meta_data, parent=self) 

    def create_dispatcher(self):
        self._set_dispatcher(self.dispatcher_creator.create_gui_element_dispatcher(self.__automator.dispatcher, self.setu_id))

    @abc.abstractmethod
    def find_if_not_found(self):
        pass

    def get_lmd(self):
        return self.__lmd

    def is_found(self):
        return self.__found

    def set_found_with(self, locator_type, locator_value):
        self.__found = True
        self.__located_by = locator_type, locator_value

    def get_found_with(self):
        return self.__located_by


class ElementConfig:

    def __init__(self, automator):
        self.__settings = {
            GuiActionConfigType.CHECK_TYPE: True,
            GuiActionConfigType.CHECK_PRE_STATE : True,
            GuiActionConfigType.CHECK_POST_STATE : True,
    }

    @property
    def settings(self):
        return self.__settings

    def configure(self, settings):
        self.__settings.update(settings)

    def _should_check_type(self):
        return self.settings[GuiActionConfigType.CHECK_TYPE]

    def _should_check_pre_state(self):
        return self.settings[GuiActionConfigType.CHECK_PRE_STATE]

    def _should_check_post_state(self):
        return self.settings[GuiActionConfigType.CHECK_POST_STATE]






import abc

from .container_conditions import GuiElementContainerConditions
from .conditions import *


class ElementContainer(metaclass=abc.ABCMeta):
    def __init__(self, config, obj_name=""):
        self.__config = config
        self.__obj_name = obj_name
        self.element_map = {}
        self.melement_map = {}
        self.__container_conditions = GuiElementContainerConditions(self)

    @property
    def max_wait_time(self):
        return self.config.get_arjuna_option_value("guiauto.max.wait").as_int()

    @property
    def config(self):
        return self.__config

    def _add_element(self, setu_id, element):
        self.element_map[setu_id] = element

    def _add_multielement(self, setu_id, melement):
        self.melement_map[setu_id] = melement

    def get_element_for_setu_id(self,id):
        return self.element_map[id]

    def get_multielement_for_setu_id(self,id):
        return self.melement_map[id]

    @abc.abstractmethod
    def _create_element_flat_or_nested(self, lmd):
        pass

    @abc.abstractmethod
    def _create_multielement_flat_or_nested(self, lmd):
        pass

    def define_element(self, lmd):
        return self._create_element_flat_or_nested(lmd)

    def define_multielement(self, lmd):
        return self._create_multielement_flat_or_nested(lmd)

    def define_dropdown(self, lmd, option_container_lmd=None, option_lmd=None):
        from arjuna.interact.gui.auto.impl.element.dropdown import GuiWebSelect
        return GuiWebSelect(self, lmd, option_container_lmd=option_container_lmd, option_lmd=option_lmd)

    def define_radiogroup(self, lmd):
        from arjuna.interact.gui.auto.impl.element.radio_group import GuiWebRadioGroup
        return GuiWebRadioGroup(self, lmd)

    def _find(self, dispatcher_call, gui_element, context="ELEMENT"):
        found = False
        js_call_name = context == "ELEMENT" and "find_element_with_js" or "find_multielement_with_js"
        js_call = getattr(self, js_call_name)
        locators = gui_element.get_lmd().locators
        if context != "ELEMENT":
            if "POINT" in {l.ltype.name for l in locators}:
                raise ConditionException("With.POINT can be used only with GuiElement.")

        for locator in locators: 
            try:
                if locator.ltype.name == "POINT":
                    # Assumption here is that this container is automator.
                    instance_count, dispatcher = js_call("return document.elementFromPoint({}, {})".format(*locator.lvalue))
                elif locator.ltype.name == "JAVASCRIPT":
                    instance_count, dispatcher = js_call(locator.lvalue)
                else:
                    instance_count, dispatcher = dispatcher_call(locator.ltype.name, locator.lvalue)
                return locator.ltype.name, locator.lvalue, instance_count, dispatcher
            except Exception as e:
                print(e)
                import traceback
                traceback.print_exc()
                continue
        if not found:
            raise Exception("Could not locate elements with locator(s): {}".format(gui_element.get_lmd()))

    def wait_until_element_found(self, gui_element):
        return self.__container_conditions.PresenceOfElement(gui_element).wait(max_wait_time=self.max_wait_time)

    def wait_until_multielement_found(self, gui_element):
        return self.__container_conditions.PresenceOfMultiElement(gui_element).wait(max_wait_time=self.max_wait_time)

    def find_multielement(self, gui_element):
        locator_type, locator_value, instance_count, dispatcher = self.wait_until_multielement_found(gui_element)
        if instance_count == 0:
            raise Exception("MultiElement could not be found with any of the provided locators.")
        gui_element.set_found_with(locator_type, locator_value)
        gui_element.dispatcher = dispatcher
        gui_element.set_instance_count(instance_count)

    def find_element(self, gui_element):
        locator_type, locator_value, instance_count, dispatcher = self.wait_until_element_found(gui_element)
        gui_element.set_found_with(locator_type, locator_value)
        gui_element.dispatcher = dispatcher



import os

from arjuna.tpi.guiauto.helpers import With
from arjuna.tpi.enums import ArjunaOption

class BaseGui:

    def __init__(self, automator, ns_dir, label=None, def_file_name=None, parent=None, register=True, multi_context=False):
        super().__init__()
        self.__automator = automator
        self.__children_map = dict()
        self.__gui_registered = False
        self.__parent = parent
        self.__impl_gui = None

        self.__label = None
        self.__def_file_name = None

        if label:
            self.set_label(label)
        else:
            self.set_label(self.__class__.__name__)

        if def_file_name:
            self.set_def_file_name(def_file_name)
        else:
            self.set_def_file_name("{}.gns".format(self.label))
        
        ns_root_dir = self.config.get_arjuna_option_value(ArjunaOption.GUIAUTO_NAMESPACE_DIR).as_str()
        self.__def_file_path = os.path.join(ns_root_dir, ns_dir, self.def_file_name)

        if register:
            self._register()

    @property
    def impl_gui(self):
        return self.__impl_gui

    @property
    def automator(self):
        return self.__automator

    @property
    def impl_automator(self):
        return self.automator.impl_automator

    @property
    def config(self):
        return self.automator.config

    @property
    def test_session(self):
        return self.automator.test_session

    @property
    def context(self):
        return self.automator.context

    @property
    def label(self):
        return self.__label

    @property
    def def_file_name(self):
        return self.__def_file_name

    @property
    def def_file_path(self):
        return self.__def_file_path

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def qual_name(self):
        return self.__class__.__qualname__

    def __check_reg_status(self):
        if self.__gui_registered:
            raise Exception("Attempt to change Gui critical attribute post registration with Setu.")

    def is_gui(self):
        return True

    def set_label(self, label):
        self.__check_reg_status()
        self.__label = label

    def set_def_file_name(self, name):
        self.__check_reg_status()
        self.__def_file_name = name

    def _register(self):
        if self.__gui_registered:
            raise Exception("Attempt to re-register Gui with Setu.")

        if not self.__parent:
            self.__impl_gui = self.test_session.define_gui(self.automator, label=self.label, name=self.name, qual_name=self.qual_name, def_file_path=self.def_file_path)
        else:
            self.__impl_gui = self.__parent.impl_gui.define_gui(self.automator, label=self.label, name=self.name, qual_name=self.qual_name, def_file_path=self.def_file_path)

        if self.__parent:
            self.__parent.add_child(self.__label, self)

        self.__load()

    def add_child(self, label, gui):
        self.__children_map[label.lower()] = gui

    def get_child(self, label):
        if label:
            if label.lower() in self.__children_map:
                return self.__children_map[label.lower()]
            else:
                raise Exception("No child Gui with label: {} defined.".format(label))
        else:
            raise Exception("Child Gui label is None.")

    def reach_until(self):
        # Children can override and write any necessary loading instructions
        pass

    def validate_readiness(self):
        pass

    def __load(self):
        try:
            self.validate_readiness()
        except:
            try:
                self.reach_until()
                self.validate_readiness()
            except Exception as e:
                raise Exception(
                    "UI [{}] with SetuId [{}] did not load as expected. Error: {}.",
                    self.__class__.__name__,
                    self.get_setu_id(),
                    str(e)
                )

    def __convert_to_with(self, input):
        w = None
        if isinstance(input, With):
            w = input
        elif type(input) is str:
            w = With.gns_name(input)
        else:
            raise Exception("A With object or name of element is expected as argument.")
        return w

    def element(self, name_or_with_obj):
        return self.impl_gui.define_element(self.__convert_to_with(name_or_with_obj))

    def multi_element(self, name_or_with_obj):
        return self.impl_gui.define_multi_element(self.__convert_to_with(name_or_with_obj))

    def dropdown(self, name_or_with_obj):
        return self.impl_gui.define_dropdown(self.__convert_to_with(name_or_with_obj))

    def radio_group(self, name_or_with_obj):
        return self.impl_gui.define_radio_group(self.__convert_to_with(name_or_with_obj))

    def alert(self, name_or_with_obj):
        return self.impl_gui.define_alert(self.__convert_to_with(name_or_with_obj))

    def frame(self, name_or_with_obj):
        return self.impl_gui.define_frame(self.__convert_to_with(name_or_with_obj))

    def child_window(self, name_or_with_obj):
        return self.impl_gui.define_child_window(self.__convert_to_with(name_or_with_obj))

    @property
    def main_window(self):
        return self.impl_gui.define_main_window()

    @property
    def dom_root(self):
        return self.impl_gui.dom_root

    @property
    def browser(self):
        return self.impl_gui.browser